@@ -57,7 +57,15 @@
 #define BLUEZ_MANAGER_INTERFACE	"org.bluez.Manager"
 #define BLUEZ_ADAPTER_INTERFACE	"org.bluez.Adapter"
 #define BLUEZ_DEVICE_INTERFACE	"org.bluez.Device"
-#define BLUEZ_INPUT_INTERFACE	"org.bluez.Input"
+
+static char * connectable_interfaces[] = {
+	"org.bluez.Audio",
+	"org.bluez.Input"
+};
+
+/* Keep in sync with above */
+#define BLUEZ_INPUT_INTERFACE	(connectable_interfaces[1])
+#define BLUEZ_AUDIO_INTERFACE (connectable_interfaces[0])
 
 static DBusGConnection *connection = NULL;
 static BluetoothClient *bluetooth_client = NULL;
@@ -288,12 +296,111 @@ static gboolean compare_path(GtkTreeStore *store,
 	return found;
 }
 
-static gboolean get_iter_from_path(GtkTreeStore *store,
-					GtkTreeIter *iter, const char *path)
+static gboolean
+get_iter_from_path (GtkTreeStore *store,
+		    GtkTreeIter *iter, const char *path)
 {
 	return iter_search(store, iter, NULL, compare_path, (gpointer) path);
 }
 
+static void
+device_services_changed (DBusGProxy *iface, const char *property,
+			 GValue *value, gpointer user_data)
+{
+	BluetoothClient *client = user_data;
+	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	GtkTreeIter iter;
+	GtkTreePath *tree_path;
+	GHashTable *table;
+	const char *path;
+	gboolean is_connected;
+
+	if (g_str_equal (property, "Connected") != FALSE) {
+		is_connected = g_value_get_boolean(value);
+	} else if (g_str_equal (property, "State") != FALSE) {
+		is_connected = (g_strcmp0(g_value_get_string (value), "connected") == 0);
+	} else
+		return;
+
+	path = dbus_g_proxy_get_path (iface);
+	if (get_iter_from_path (priv->store, &iter, path) == FALSE)
+		return;
+
+	gtk_tree_model_get(GTK_TREE_MODEL (priv->store), &iter,
+			   BLUETOOTH_COLUMN_SERVICES, &table,
+			   -1);
+
+	g_hash_table_insert (table,
+			     (gpointer) dbus_g_proxy_get_interface (iface),
+			     GINT_TO_POINTER (is_connected));
+
+	tree_path = gtk_tree_model_get_path (GTK_TREE_MODEL (priv->store), &iter);
+	gtk_tree_model_row_changed (GTK_TREE_MODEL (priv->store), tree_path, &iter);
+	gtk_tree_path_free (tree_path);
+}
+
+static GHashTable *
+device_list_nodes (DBusGProxy *device, BluetoothClient *client, gboolean connect_signal)
+{
+	GHashTable *table;
+	guint i;
+
+	if (device == NULL)
+		return NULL;
+
+	table = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
+
+	for (i = 0; i < G_N_ELEMENTS (connectable_interfaces); i++) {
+		DBusGProxy *iface;
+		GHashTable *props;
+
+		/* Don't add the input interface for devices that already have
+		 * audio stuff */
+		if (g_str_equal (connectable_interfaces[i], BLUEZ_INPUT_INTERFACE)
+		    && g_hash_table_size (table) > 0)
+			continue;
+
+		iface = dbus_g_proxy_new_from_proxy (device, connectable_interfaces[i],
+						     NULL);
+		if (dbus_g_proxy_call (iface, "GetProperties", NULL,
+				       G_TYPE_INVALID, dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE), &props,
+				       G_TYPE_INVALID) != FALSE) {
+			GValue *value;
+			gboolean is_connected;
+
+			value = g_hash_table_lookup(props, "Connected");
+			if (value != NULL) {
+				is_connected = g_value_get_boolean(value);
+			} else {
+				const char *str = "disconnected";
+				value = g_hash_table_lookup(props, "State");
+				if (value != NULL)
+					str = g_value_get_string(value);
+
+				is_connected = (g_strcmp0(str, "connected") == 0);
+			}
+
+			g_hash_table_insert (table,
+					     connectable_interfaces[i],
+					     GINT_TO_POINTER (is_connected));
+
+			if (connect_signal != FALSE) {
+				dbus_g_proxy_add_signal(iface, "PropertyChanged",
+							G_TYPE_STRING, G_TYPE_VALUE, G_TYPE_INVALID);
+				dbus_g_proxy_connect_signal(iface, "PropertyChanged",
+							    G_CALLBACK(device_services_changed), client, NULL);
+			}
+		}
+	}
+
+	if (g_hash_table_size (table) == 0) {
+		g_hash_table_destroy (table);
+		return NULL;
+	}
+
+	return table;
+}
+
 static void device_changed(DBusGProxy *device, const char *property,
 					GValue *value, gpointer user_data)
 {
@@ -336,6 +443,12 @@ static void device_changed(DBusGProxy *device, const char *property,
 
 		gtk_tree_store_set(priv->store, &iter,
 				BLUETOOTH_COLUMN_CONNECTED, connected, -1);
+	} else if (g_str_equal (property, "UUIDs") == TRUE) {
+		GHashTable *services;
+
+		services = device_list_nodes (device, client, TRUE);
+		gtk_tree_store_set(priv->store, &iter,
+				   BLUETOOTH_COLUMN_SERVICES, services, -1);
 	}
 }
 
@@ -347,6 +460,7 @@ static void add_device(DBusGProxy *adapter, GtkTreeIter *parent,
 	DBusGProxy *device;
 	GValue *value;
 	const gchar *address, *alias, *name, *icon;
+	GHashTable *services;
 	gboolean paired, trusted, connected;
 	guint type;
 	gint rssi;
@@ -390,15 +504,9 @@ static void add_device(DBusGProxy *adapter, GtkTreeIter *parent,
 		value = g_hash_table_lookup(hash, "Connected");
 		connected = value ? g_value_get_boolean(value) : FALSE;
 	} else {
-		address = NULL;
-		alias = NULL;
-		name = NULL;
-		type = BLUETOOTH_TYPE_ANY;
-		icon = NULL;
-		rssi = 0;
-		paired = FALSE;
-		trusted = FALSE;
-		connected = FALSE;
+		if (device)
+			g_object_unref (device);
+		return;
 	}
 
 	cont = gtk_tree_model_iter_children(GTK_TREE_MODEL(priv->store),
@@ -417,14 +525,19 @@ static void add_device(DBusGProxy *adapter, GtkTreeIter *parent,
 					BLUETOOTH_COLUMN_NAME, name,
 					BLUETOOTH_COLUMN_TYPE, type,
 					BLUETOOTH_COLUMN_ICON, icon,
-					BLUETOOTH_COLUMN_RSSI, rssi, -1);
+					BLUETOOTH_COLUMN_RSSI, rssi,
+					-1);
 
 			if (device != NULL) {
+				services = device_list_nodes (device, client, FALSE);
+
 				gtk_tree_store_set(priv->store, &iter,
 					BLUETOOTH_COLUMN_PROXY, device,
 					BLUETOOTH_COLUMN_CONNECTED, connected,
 					BLUETOOTH_COLUMN_TRUSTED, trusted,
-					BLUETOOTH_COLUMN_PAIRED, paired, -1);
+					BLUETOOTH_COLUMN_PAIRED, paired,
+					BLUETOOTH_COLUMN_SERVICES, services,
+					-1);
 			}
 
 			goto done;
@@ -434,6 +547,8 @@ static void add_device(DBusGProxy *adapter, GtkTreeIter *parent,
 		cont = gtk_tree_model_iter_next(GTK_TREE_MODEL(priv->store), &iter);
 	}
 
+	services = device_list_nodes (device, client, TRUE);
+
 	gtk_tree_store_insert_with_values(priv->store, &iter, parent, -1,
 				BLUETOOTH_COLUMN_PROXY, device,
 				BLUETOOTH_COLUMN_ADDRESS, address,
@@ -444,7 +559,9 @@ static void add_device(DBusGProxy *adapter, GtkTreeIter *parent,
 				BLUETOOTH_COLUMN_RSSI, rssi,
 				BLUETOOTH_COLUMN_PAIRED, paired,
 				BLUETOOTH_COLUMN_TRUSTED, trusted,
-				BLUETOOTH_COLUMN_CONNECTED, connected, -1);
+				BLUETOOTH_COLUMN_CONNECTED, connected,
+				BLUETOOTH_COLUMN_SERVICES, services,
+				-1);
 
 done:
 	if (device != NULL) {
@@ -452,7 +569,6 @@ done:
 				G_TYPE_STRING, G_TYPE_VALUE, G_TYPE_INVALID);
 		dbus_g_proxy_connect_signal(device, "PropertyChanged",
 				G_CALLBACK(device_changed), client, NULL);
-
 		g_object_unref(device);
 	}
 }
@@ -794,7 +910,8 @@ static void bluetooth_client_init(BluetoothClient *client)
 					 G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING,
 					 G_TYPE_UINT, G_TYPE_STRING, G_TYPE_INT,
 					 G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN,
-					 G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
+					 G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN,
+					 G_TYPE_HASH_TABLE);
 
 	priv->dbus = dbus_g_proxy_new_for_name(connection, DBUS_SERVICE_DBUS,
 				DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
@@ -934,27 +1051,33 @@ BluetoothClient *bluetooth_client_new(void)
 	return bluetooth_client;
 }
 
-GtkTreeModel *bluetooth_client_get_model(BluetoothClient *client)
+GtkTreeModel *bluetooth_client_get_model (BluetoothClient *client)
 {
-	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	BluetoothClientPrivate *priv;
 	GtkTreeModel *model;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), NULL);
+
 	DBG("client %p", client);
 
+	priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	model = g_object_ref(priv->store);
 
 	return model;
 }
 
-GtkTreeModel *bluetooth_client_get_filter_model(BluetoothClient *client,
-				GtkTreeModelFilterVisibleFunc func,
-				gpointer data, GDestroyNotify destroy)
+GtkTreeModel *bluetooth_client_get_filter_model (BluetoothClient *client,
+						 GtkTreeModelFilterVisibleFunc func,
+						 gpointer data, GDestroyNotify destroy)
 {
-	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	BluetoothClientPrivate *priv;
 	GtkTreeModel *model;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), NULL);
+
 	DBG("client %p", client);
 
+	priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	model = gtk_tree_model_filter_new(GTK_TREE_MODEL(priv->store), NULL);
 
 	gtk_tree_model_filter_set_visible_func(GTK_TREE_MODEL_FILTER(model),
@@ -982,25 +1105,28 @@ static gboolean adapter_filter(GtkTreeModel *model,
 	return active;
 }
 
-GtkTreeModel *bluetooth_client_get_adapter_model(BluetoothClient *client)
+GtkTreeModel *bluetooth_client_get_adapter_model (BluetoothClient *client)
 {
 	DBG("client %p", client);
 
-	return bluetooth_client_get_filter_model(client, adapter_filter,
-								NULL, NULL);
+	return bluetooth_client_get_filter_model (client, adapter_filter,
+						  NULL, NULL);
 }
 
-GtkTreeModel *bluetooth_client_get_device_model(BluetoothClient *client,
-							DBusGProxy *adapter)
+GtkTreeModel *bluetooth_client_get_device_model (BluetoothClient *client,
+						 DBusGProxy *adapter)
 {
-	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	BluetoothClientPrivate *priv;
 	GtkTreeModel *model;
 	GtkTreePath *path;
 	GtkTreeIter iter;
 	gboolean cont, found = FALSE;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), NULL);
+
 	DBG("client %p", client);
 
+	priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	cont = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(priv->store),
 									&iter);
 
@@ -1045,6 +1171,8 @@ GtkTreeModel *bluetooth_client_get_device_filter_model(BluetoothClient *client,
 {
 	GtkTreeModel *model;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), NULL);
+
 	DBG("client %p", client);
 
 	model = bluetooth_client_get_device_model(client, adapter);
@@ -1061,6 +1189,8 @@ DBusGProxy *bluetooth_client_get_default_adapter(BluetoothClient *client)
 	GtkTreeIter iter;
 	gboolean cont;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), NULL);
+
 	DBG("client %p", client);
 
 	cont = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(priv->store),
@@ -1090,6 +1220,8 @@ gboolean bluetooth_client_start_discovery(BluetoothClient *client)
 {
 	DBusGProxy *adapter;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+
 	DBG("client %p", client);
 
 	adapter = bluetooth_client_get_default_adapter(client);
@@ -1107,6 +1239,8 @@ gboolean bluetooth_client_stop_discovery(BluetoothClient *client)
 {
 	DBusGProxy *adapter;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+
 	DBG("client %p", client);
 
 	adapter = bluetooth_client_get_default_adapter(client);
@@ -1154,6 +1288,9 @@ gboolean bluetooth_client_create_device(BluetoothClient *client,
 	DBusGProxy *adapter;
 	DBusGProxyCall *call;
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+	g_return_val_if_fail (address != NULL, FALSE);
+
 	DBG("client %p", client);
 
 	adapter = bluetooth_client_get_default_adapter(client);
@@ -1190,6 +1327,9 @@ gboolean bluetooth_client_set_trusted(BluetoothClient *client,
 	DBusGProxy *proxy;
 	GValue value = { 0 };
 
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+	g_return_val_if_fail (device != NULL, FALSE);
+
 	DBG("client %p", client);
 
 	proxy = dbus_g_proxy_new_from_proxy(priv->manager,
@@ -1215,10 +1355,12 @@ gboolean bluetooth_client_set_trusted(BluetoothClient *client,
 typedef struct {
 	BluetoothClientConnectFunc func;
 	gpointer data;
+	/* used for disconnect */
+	GList *services;
 } ConnectData;
 
-static void connect_input_callback(DBusGProxy *proxy,
-					DBusGProxyCall *call, void *user_data)
+static void connect_callback(DBusGProxy *proxy,
+			     DBusGProxyCall *call, void *user_data)
 {
 	ConnectData *conndata = user_data;
 	GError *error = NULL;
@@ -1234,34 +1376,183 @@ static void connect_input_callback(DBusGProxy *proxy,
 	g_object_unref(proxy);
 }
 
-gboolean bluetooth_client_connect_input(BluetoothClient *client,
-				const char *device,
-				BluetoothClientConnectFunc func, gpointer data)
+gboolean bluetooth_client_connect_service(BluetoothClient *client,
+					  const char *device,
+					  BluetoothClientConnectFunc func,
+					  gpointer data)
 {
 	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	ConnectData *conndata;
 	DBusGProxy *proxy;
-	DBusGProxyCall *call;
+	GHashTable *table;
+	GtkTreeIter iter;
+	const char *iface_name;
+	guint i;
+
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+	g_return_val_if_fail (device != NULL, FALSE);
 
 	DBG("client %p", client);
 
-	proxy = dbus_g_proxy_new_from_proxy(priv->manager,
-						BLUEZ_INPUT_INTERFACE, device);
-	if (proxy == NULL)
+	if (get_iter_from_path (priv->store, &iter, device) == FALSE)
 		return FALSE;
 
-	conndata = g_try_new0(ConnectData, 1);
-	if (conndata == NULL) {
-		g_object_unref(proxy);
+	gtk_tree_model_get(GTK_TREE_MODEL (priv->store), &iter,
+			   BLUETOOTH_COLUMN_SERVICES, &table,
+			   -1);
+	if (table == NULL)
+		return FALSE;
+
+	conndata = g_new0 (ConnectData, 1);
+
+	iface_name = NULL;
+	for (i = 0; i < G_N_ELEMENTS (connectable_interfaces); i++) {
+		if (g_hash_table_lookup_extended (table, connectable_interfaces[i], NULL, NULL) != FALSE) {
+			iface_name = connectable_interfaces[i];
+			break;
+		}
+	}
+
+	if (iface_name == NULL) {
+		g_printerr("No supported services on the '%s' device\n", device);
+		g_free (conndata);
+		return FALSE;
+	}
+
+	proxy = dbus_g_proxy_new_from_proxy(priv->manager,
+						iface_name, device);
+	if (proxy == NULL) {
+		g_free (conndata);
 		return FALSE;
 	}
 
 	conndata->func = func;
 	conndata->data = data;
 
-	call = dbus_g_proxy_begin_call(proxy, "Connect",
-				connect_input_callback, conndata, g_free,
-							G_TYPE_INVALID);
+	dbus_g_proxy_begin_call(proxy, "Connect",
+				connect_callback, conndata, g_free,
+				G_TYPE_INVALID);
+
+	g_object_unref (proxy);
 
 	return TRUE;
 }
+
+static void disconnect_callback(DBusGProxy *proxy,
+				DBusGProxyCall *call,
+				void *user_data)
+{
+	ConnectData *conndata = user_data;
+
+	dbus_g_proxy_end_call(proxy, call, NULL, G_TYPE_INVALID);
+
+	if (conndata->services != NULL) {
+		DBusGProxy *service;
+
+		service = dbus_g_proxy_new_from_proxy (proxy,
+						       conndata->services->data, NULL);
+
+		conndata->services = g_list_remove (conndata->services, conndata->services->data);
+
+		dbus_g_proxy_begin_call(service, "Disconnect",
+					disconnect_callback, conndata, NULL,
+					G_TYPE_INVALID);
+
+		g_object_unref (proxy);
+
+		return;
+	}
+
+	if (conndata->func)
+		conndata->func(conndata->data);
+
+	g_object_unref(proxy);
+	g_free (conndata);
+}
+
+static int
+service_to_index (const char *service)
+{
+	guint i;
+
+	g_return_val_if_fail (service != NULL, -1);
+
+	for (i = 0; i < G_N_ELEMENTS (connectable_interfaces); i++) {
+		if (g_str_equal (connectable_interfaces[i], service) != FALSE)
+			return i;
+	}
+
+	g_assert_not_reached ();
+
+	return -1;
+}
+
+static int
+rev_sort_services (const char *servicea, const char *serviceb)
+{
+	int a, b;
+
+	a = service_to_index (servicea);
+	b = service_to_index (serviceb);
+
+	if (a < b)
+		return 1;
+	if (a > b)
+		return -1;
+	return 0;
+}
+
+gboolean bluetooth_client_disconnect_service (BluetoothClient *client,
+					      const char *device,
+					      BluetoothClientConnectFunc func,
+					      gpointer data)
+{
+	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
+	ConnectData *conndata;
+	DBusGProxy *proxy;
+	GHashTable *table;
+	GtkTreeIter iter;
+
+	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
+	g_return_val_if_fail (device != NULL, FALSE);
+
+	DBG("client %p", client);
+
+	if (get_iter_from_path (priv->store, &iter, device) == FALSE)
+		return FALSE;
+
+	gtk_tree_model_get(GTK_TREE_MODEL (priv->store), &iter,
+			   BLUETOOTH_COLUMN_PROXY, &proxy,
+			   BLUETOOTH_COLUMN_SERVICES, &table,
+			   -1);
+	if (proxy == NULL)
+		return FALSE;
+
+	conndata = g_new0 (ConnectData, 1);
+
+	conndata->func = func;
+	conndata->data = data;
+
+	if (table == NULL) {
+		dbus_g_proxy_begin_call(proxy, "Disconnect",
+					disconnect_callback, conndata, NULL,
+					G_TYPE_INVALID);
+	} else {
+		DBusGProxy *service;
+
+		conndata->services = g_hash_table_get_keys (table);
+		conndata->services = g_list_sort (conndata->services, (GCompareFunc) rev_sort_services);
+
+		service = dbus_g_proxy_new_from_proxy (priv->manager,
+						       conndata->services->data, device);
+
+		conndata->services = g_list_remove (conndata->services, conndata->services->data);
+
+		dbus_g_proxy_begin_call(service, "Disconnect",
+					disconnect_callback, conndata, NULL,
+					G_TYPE_INVALID);
+	}
+
+	return TRUE;
+}
+
